#pragma once
#include "Shader.h"



Shader::Shader(const std::string& vertexShader,const std::string& fragmentShader)
	: programHandle(0)
	, _vertexHandle(0)
	, _fragmentHandle(0)
	, _geometryHandle(0)
{
	programHandle = glCreateProgram();
	if (programHandle == 0)
	{
		cerr<<"ERROR: failed to create shader program"<<endl;
	}
	loadShader(vertexShader, GL_VERTEX_SHADER, _vertexHandle);
	loadShader(fragmentShader, GL_FRAGMENT_SHADER, _fragmentHandle);

	link();
}

Shader::Shader(std::vector<ShaderStruct>& shaders)
	: programHandle(0)
	, _vertexHandle(0)
	, _fragmentHandle(0)
	, _geometryHandle(0)
{
	programHandle = glCreateProgram();
	if (programHandle == 0)
	{
		cerr << "ERROR: failed to create shader program"<<endl;
	}

	for(auto &s : shaders)
	{
		switch (s._type)
		{
			case VERT: loadShader(s._shader, GL_VERTEX_SHADER, _vertexHandle); break;
			case FRAG: loadShader(s._shader, GL_FRAGMENT_SHADER, _fragmentHandle); break;
			case GEOM: loadShader(s._shader, GL_GEOMETRY_SHADER, _geometryHandle); break;
		}
	}

	link();
}


Shader::~Shader()
{
	glDeleteProgram(programHandle);
	glDeleteShader(_vertexHandle);
	glDeleteShader(_fragmentHandle);
	glDeleteShader(_geometryHandle);
}

void Shader::useShader() const
{
	glUseProgram(programHandle);
}

void Shader::loadShader(const std::string& shader,GLenum shaderType, GLuint& handle)
{
	ifstream shaderFile(shader);

	if (shaderFile.good())
	{
		auto code = string(istreambuf_iterator<char>(shaderFile), istreambuf_iterator<char>());
		shaderFile.close();

		handle = glCreateShader(shaderType);
		if (handle == 0)
		{
			cerr<<"ERROR: failed to create shader "<<endl;
		}

		auto code_ptr = code.c_str();
		glShaderSource(handle, 1, &code_ptr, nullptr);
		glCompileShader(handle);

		GLint succeeded;
		glGetShaderiv(handle, GL_COMPILE_STATUS, &succeeded);
		if (!succeeded || !glIsShader(handle))
		{
			GLint logSize;
			glGetShaderiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			auto msg = make_unique<char[]>(logSize);
			glGetShaderInfoLog(handle, logSize, nullptr, msg.get());

			string retMsg = msg.get();
			msg.reset(nullptr);
			cerr<<retMsg<<endl;
		}
	}
	else
	{
		cerr<<"ERROR: failed to open shader file"<<endl;
	}
}

void Shader::setUniform(std::string sName, float* fValues, int iCount)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniform1fv(iLoc, iCount, fValues);
}

void Shader::setUniform(std::string sName, const float fValue)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniform1fv(iLoc, 1, &fValue);
}

void Shader::setUniform(std::string sName, int* iValues, int iCount)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniform1iv(iLoc, iCount, iValues);
}

void Shader::setUniform(std::string sName, const int iValue)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniform1i(iLoc, iValue);
}

void Shader::setUniform(std::string sName, const glm::vec3 vVector)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniform3fv(iLoc, 1, value_ptr(vVector));
}

void Shader::setUniform(std::string sName, const glm::mat4 mMatrix)
{
	int iLoc = glGetUniformLocation(programHandle, sName.c_str());
	glUniformMatrix4fv(iLoc, 1, GL_FALSE, value_ptr(mMatrix));
}

void Shader::link()
{
	glAttachShader(programHandle, _vertexHandle);
	// glAttachShader(programHandle, _geometryHandle); throws error because not used.
	glAttachShader(programHandle, _fragmentHandle);

	glLinkProgram(programHandle);
	GLint succeeded;
	glGetProgramiv(programHandle, GL_LINK_STATUS, &succeeded);
	if (!succeeded)
	{
		GLint logSize;
		glGetProgramiv(programHandle, GL_INFO_LOG_LENGTH, &logSize);

		auto msg = make_unique<char[]>(logSize);
		glGetProgramInfoLog(programHandle, logSize, nullptr, msg.get());

		string retMsg = msg.get();
		msg.reset(nullptr);
		cerr<<retMsg<<endl;
	}
}